/*
 cc -Ae -D_KERNEL -D_KERNEL_BUILD -DBPFMOD_DEBUG -DCONVERGED_IO -c bpfmod.c
              Copyright (C) 1998 Hewlett-Packard Company
                         ALL RIGHTS RESERVED.
 
  The enclosed software and documentation includes copyrighted works
  of Hewlett-Packard Co. For as long as you comply with the following
  limitations, you are hereby authorized to (i) use, reproduce, and
  modify the software and documentation, and to (ii) distribute the
  software and documentation, including modifications, for
  non-commercial purposes only.
      
  1.  The enclosed software and documentation is made available at no
      charge in order to advance the general development of
      high-performance networking products.
 
  2.  You may not delete any copyright notices contained in the
      software or documentation. All hard copies, and copies in source
      code or object code form, of the software or documentation
      (including modifications) must contain at least one of the
      copyright notices.
 
  3.  The enclosed software and documentation has not been subjected
      to testing and quality control and is not a Hewlett-Packard Co.
      product. At a future time, Hewlett-Packard Co. may or may not
      offer a version of the software and documentation as a product.
  
  4.  THE SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS".
      HEWLETT-PACKARD COMPANY DOES NOT WARRANT THAT THE USE,
      REPRODUCTION, MODIFICATION OR DISTRIBUTION OF THE SOFTWARE OR
      DOCUMENTATION WILL NOT INFRINGE A THIRD PARTY'S INTELLECTUAL
      PROPERTY RIGHTS. HP DOES NOT WARRANT THAT THE SOFTWARE OR
      DOCUMENTATION IS ERROR FREE. HP DISCLAIMS ALL WARRANTIES,
      EXPRESS AND IMPLIED, WITH REGARD TO THE SOFTWARE AND THE
      DOCUMENTATION. HP SPECIFICALLY DISCLAIMS ALL WARRANTIES OF
      MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  
  5.  HEWLETT-PACKARD COMPANY WILL NOT IN ANY EVENT BE LIABLE FOR ANY
      DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
      (INCLUDING LOST PROFITS) RELATED TO ANY USE, REPRODUCTION,
      MODIFICATION, OR DISTRIBUTION OF THE SOFTWARE OR DOCUMENTATION.
 */

/* Much of this code is leveraged liberally from the BSD BPF code, so
 * we include the following:
 *
 * Copyright (c) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from the Stanford/CMU enet packet filter,
 * (net/enet.c) distributed as part of 4.3BSD, and code contributed
 * to Berkeley by Steven McCanne and Van Jacobson both of Lawrence 
 * Berkeley Laboratory.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)bpf.c	7.5 (Berkeley) 7/15/91
 */

#if  defined(MODULEID) && !defined(lint)
const static char rcsid[] = "@(#) bpfmod.c experimental 0.10 06/03/1999";
#endif

#ifdef OSDEBUG
#define BPFMOD_DEBUG
#endif /* OSDEBUG */

/*
 bpfmod (Berkeley Packet Filter Module)


 This module implements the filtering and packet encapsulation
 portions of the Berkeley Packet Filter. When pushed onto the stack,
 it will pass-on only those packets matching the user-supplied
 filter. Other packets will be discarded. If not in "immediate mode"
 packets will be concatenated into larger buffers to minimize the
 overhead of reading those packets into the application.

 raj 10/98 */

#include "../h/types.h"
#include "../h/param.h"
#include "../h/errno.h"

#include "../h/stropts.h"

#ifdef nodef
#include "../streams/str_kernel.h"
#endif

#ifdef _HPUX_SOURCE
#include "../h/_types.h"
#include "../h/time.h"
#include "../h/resource.h"
#include "../h/netdiag1.h"
/* it would seem that 10.20 netdiag1.h does not contain an
   ms_gettimeofday() declaration. I wonder what happens on 11.X? */
extern struct timeval ms_gettimeofday(void);
#include "../h/malloc.h"
#endif

#include "../h/dlpi.h"
#include "../h/dlpi_ext.h"
#include "../h/stream.h"

/* this is an ugly hack to get the struct ifreq into the compiler
   without having to include <net/if.h> - there seems to be no h/if.h
   file...this is only really needed for the GIF/SIF ioctls */
#include "../h/socket.h"
struct  ifreq {
#define IFNAMSIZ        16
        char    ifr_name[IFNAMSIZ];             /* if name, e.g. "en0" */
        union {
                struct  sockaddr ifru_addr;
                struct  sockaddr ifru_dstaddr;
                struct  sockaddr ifru_broadaddr;
                short   ifru_flags;
                int     ifru_metric;
                caddr_t ifru_data;
        } ifr_ifru;
#define ifr_addr        ifr_ifru.ifru_addr      /* address */
#define ifr_dstaddr     ifr_ifru.ifru_dstaddr   /* other end of p-to-p link */
#define ifr_broadaddr   ifr_ifru.ifru_broadaddr /* broadcast address */
#define ifr_flags       ifr_ifru.ifru_flags     /* flags */
#define ifr_metric      ifr_ifru.ifru_metric    /* metric */
#define ifr_data        ifr_ifru.ifru_data      /* for use by interface */
};

#include "bpfmod.h"

#include "../h/strenv.h"
/*
 * Pre-declare the open, close, service and put routines for the data
 * structures to follow...
 */

int bpfmodopen(), bpfmodclose(), bpfmodsrv(), bpfmodrput(), bpfmodwput();

static struct module_info bpfmodm_info = 
{BPFMOD_ID, "bpfmod", 0, INFPSZ, 0, 0 };

static struct qinit bpfmodrinit = 
{ bpfmodrput, NULL, bpfmodopen, bpfmodclose, NULL, &bpfmodm_info, NULL };

static struct qinit bpfmodwinit = 
{ bpfmodwput, NULL, NULL, NULL, NULL, &bpfmodm_info, NULL };

struct streamtab bpfmodinfo = { &bpfmodrinit, &bpfmodwinit };

#ifdef CONVERGED_IO
#include "../h/conf.h"

static streams_info_t bpfmod_str_info = {
    "bpfmod",                       		/* name */
    -1,                         		/* major number */
    { &bpfmodrinit, &bpfmodwinit, NULL, NULL },	/* streamtab */
    STR_IS_MODULE,		  		/* streams flags */
    SQLVL_QUEUEPAIR,            		/* sync level */
    "",                         		/* elsewhere sync name */
};
#endif /* CONVERGED_IO */      

/* Now that we have all that streams declaration stuff out of the way,
   lets get the stuff for the "meat" of bpfmod. */

/*
 * The default read buffer size is patchable.
 */
int bpf_bufsize = BPF_BUFSIZE;

/* This is a heavily gutted version of the bpf_d that has only those
   fields valid for bpfmod. this is to make some of the code
   modifications a little easier - less cuting of things and all
   that... of course, I may just modify things in their entirety. i've
   moved some fields around so that when/if I take this to a 64 bit
   kernel, there will be no holes in the datastructure raj 10/98 */

struct bpf_d {
  mblk_t          *bd_sbuf_mp;    /* mblock containing the store slot */
  struct bpf_insn *bd_filter;     /* filter code */
  mblk_t          *bd_timer_mp;   /* just a dummy mblk used for
				     streams_put timers */
  int             bd_rtout;       /* the "read" timeout in ticks */
  int             bd_timer_id;    /* is there currently a timer
				     running? */
  int             bd_slen;        /* the current number of bytes in
				     the store slot */
  int             bd_bufsize;     /* absolute length of buffers */
  
  bpf_u_int32     bd_hdrlen;      /* the computed size of the bpf
				     header. this is computed so the
				     network layer header will align
				     on a longword (4 byte) bondary,
				     which makes life easier for those
				     above. assuming, of course, that
				     the wiretapping of certain DL_
				     primitives as they go by on the
				     stream is successful */
  bpf_u_int32     bd_immediate;   /* are we in immediate mode? */
  bpf_u_int32     bd_rcount;      /* number of packets received */
  bpf_u_int32     bd_dcount;      /* number of packets dropped */
  bpf_u_int32     bd_dlt;         /* the data link type in BPFspeak
				     for the driver below us */
};

/* some systems get cranky if structures are not aligned
   nicely. PA-RISC is such a system */

#define EXTRACT_SHORT(p)\
	((u_short)\
		((u_short)*((u_char *)p+0)<<8|\
		 (u_short)*((u_char *)p+1)<<0))
#define EXTRACT_LONG(p)\
		((int32_t)*((u_char *)p+0)<<24|\
		 (int32_t)*((u_char *)p+1)<<16|\
		 (int32_t)*((u_char *)p+2)<<8|\
		 (int32_t)*((u_char *)p+3)<<0)

#ifdef _KERNEL
#define MINDEX(len, m, k) \
{ \
	len = m->b_wptr - m->b_rptr; \
	while (k >= len) { \
		k -= len; \
		m = m->b_cont; \
		if (m == 0) \
			return 0; \
		len = m->b_wptr - m->b_rptr; \
	} \
}

static int
m_xword(m, k, err)
	mblk_t *m;
	int k, *err;
{
	int len;
	u_char *cp, *np;
	mblk_t *m0;

	MINDEX(len, m, k);
	cp = m->b_rptr + k;
	if (len - k >= 4) {
		*err = 0;
		return EXTRACT_LONG(cp);
	}
	m0 = m->b_cont;
	if (m0 == NULL || (m0->b_wptr - m0->b_rptr) + len - k < 4)
		goto bad;
	*err = 0;
	np = m0->b_rptr;
	switch (len - k) {

	case 1:
		return (cp[0] << 24) | (np[0] << 16) | (np[1] << 8) | np[2];

	case 2:
		return (cp[0] << 24) | (cp[1] << 16) | (np[0] << 8) | np[1];

	default:
		return (cp[0] << 24) | (cp[1] << 16) | (cp[2] << 8) | np[0];
	}
    bad:
	*err = 1;
	return 0;
}

static int
m_xhalf(m, k, err)
	mblk_t *m;
	int k, *err;
{
	int len;
	u_char *cp;
	mblk_t *m0;

	MINDEX(len, m, k);
	cp = m->b_rptr + k;
	if (len - k >= 2) {
		*err = 0;
		return EXTRACT_SHORT(cp);
	}
	m0 = m->b_cont;
	if (m0 == 0)
		goto bad;
	*err = 0;
	return (cp[0] << 8) | ((u_char *)(m0->b_rptr))[0];
 bad:
	*err = 1;
	return 0;
}
#endif

/*
 * Execute the filter program starting at pc on the packet p
 * wirelen is the length of the original packet
 * buflen is the amount of data present
 */
u_int
bpf_filter(pc, p, wirelen, buflen)
	struct bpf_insn *pc;
	u_char *p;
	u_int wirelen;
	u_int buflen;
{
	int32_t A, X;
	int k;
	int32_t mem[BPF_MEMWORDS];

	if (pc == NULL)
		/*
		 * No filter means accept all.
		 */
		return (u_int)-1;
	A = 0;
	X = 0;
	--pc;
	while (1) {
		++pc;
		switch (pc->code) {

		default:
#ifdef _KERNEL
			return 0;
#else
			abort();
#endif			
		case BPF_RET|BPF_K:
			return (u_int)pc->k;

		case BPF_RET|BPF_A:
			return (u_int)A;

		case BPF_LD|BPF_W|BPF_ABS:
			k = pc->k;
			if (k + sizeof(int32_t) > buflen) {
#ifdef _KERNEL
				int merr;

				if (buflen != 0)
					return 0;
				A = m_xword((mblk_t *)p, k, &merr);
				if (merr != 0)
					return 0;
				continue;
#else
				return 0;
#endif
			}
			A = EXTRACT_LONG(&p[k]);
			continue;

		case BPF_LD|BPF_H|BPF_ABS:
			k = pc->k;
			if (k + sizeof(short) > buflen) {
#ifdef _KERNEL
				int merr;

				if (buflen != 0)
					return 0;
				A = m_xhalf((mblk_t *)p, k, &merr);
				continue;
#else
				return 0;
#endif
			}
			A = EXTRACT_SHORT(&p[k]);
			continue;

		case BPF_LD|BPF_B|BPF_ABS:
			k = pc->k;
			if (k >= buflen) {
#ifdef _KERNEL
				mblk_t *m;
				register int len;

				if (buflen != 0)
					return 0;
				m = (mblk_t *)p;
				MINDEX(len, m, k);
				A = ((u_char *)(m->b_rptr))[k];
				continue;
#else
				return 0;
#endif
			}
			A = p[k];
			continue;

		case BPF_LD|BPF_W|BPF_LEN:
			A = wirelen;
			continue;

		case BPF_LDX|BPF_W|BPF_LEN:
			X = wirelen;
			continue;

		case BPF_LD|BPF_W|BPF_IND:
			k = X + pc->k;
			if (k + sizeof(int32_t) > buflen) {
#ifdef _KERNEL
				int merr;

				if (buflen != 0)
					return 0;
				A = m_xword((mblk_t *)p, k, &merr);
				if (merr != 0)
					return 0;
				continue;
#else
				return 0;
#endif
			}
			A = EXTRACT_LONG(&p[k]);
			continue;

		case BPF_LD|BPF_H|BPF_IND:
			k = X + pc->k;
			if (k + sizeof(short) > buflen) {
#ifdef _KERNEL
				int merr;

				if (buflen != 0)
					return 0;
				A = m_xhalf((mblk_t *)p, k, &merr);
				if (merr != 0)
					return 0;
				continue;
#else
				return 0;
#endif
			}
			A = EXTRACT_SHORT(&p[k]);
			continue;

		case BPF_LD|BPF_B|BPF_IND:
			k = X + pc->k;
			if (k >= buflen) {
#ifdef _KERNEL
				mblk_t *m;
				int len;

				if (buflen != 0)
					return 0;
				m = (mblk_t *)p;
				MINDEX(len, m, k);
				A = ((u_char *)(m->b_rptr))[k];
				continue;
#else
				return 0;
#endif
			}
			A = p[k];
			continue;

		case BPF_LDX|BPF_MSH|BPF_B:
			k = pc->k;
			if (k >= buflen) {
#ifdef _KERNEL
				mblk_t *m;
				int len;

				if (buflen != 0)
					return 0;
				m = (mblk_t *)p;
				MINDEX(len, m, k);
				X = (((char *)(m->b_rptr))[k] & 0xf) << 2;
				continue;
#else
				return 0;
#endif
			}
			X = (p[pc->k] & 0xf) << 2;
			continue;

		case BPF_LD|BPF_IMM:
			A = pc->k;
			continue;

		case BPF_LDX|BPF_IMM:
			X = pc->k;
			continue;

		case BPF_LD|BPF_MEM:
			A = mem[pc->k];
			continue;
			
		case BPF_LDX|BPF_MEM:
			X = mem[pc->k];
			continue;

		case BPF_ST:
			mem[pc->k] = A;
			continue;

		case BPF_STX:
			mem[pc->k] = X;
			continue;

		case BPF_JMP|BPF_JA:
			pc += pc->k;
			continue;

		case BPF_JMP|BPF_JGT|BPF_K:
			pc += (A > pc->k) ? pc->jt : pc->jf;
			continue;

		case BPF_JMP|BPF_JGE|BPF_K:
			pc += (A >= pc->k) ? pc->jt : pc->jf;
			continue;

		case BPF_JMP|BPF_JEQ|BPF_K:
			pc += (A == pc->k) ? pc->jt : pc->jf;
			continue;

		case BPF_JMP|BPF_JSET|BPF_K:
			pc += (A & pc->k) ? pc->jt : pc->jf;
			continue;

		case BPF_JMP|BPF_JGT|BPF_X:
			pc += (A > X) ? pc->jt : pc->jf;
			continue;

		case BPF_JMP|BPF_JGE|BPF_X:
			pc += (A >= X) ? pc->jt : pc->jf;
			continue;

		case BPF_JMP|BPF_JEQ|BPF_X:
			pc += (A == X) ? pc->jt : pc->jf;
			continue;

		case BPF_JMP|BPF_JSET|BPF_X:
			pc += (A & X) ? pc->jt : pc->jf;
			continue;

		case BPF_ALU|BPF_ADD|BPF_X:
			A += X;
			continue;
			
		case BPF_ALU|BPF_SUB|BPF_X:
			A -= X;
			continue;
			
		case BPF_ALU|BPF_MUL|BPF_X:
			A *= X;
			continue;
			
		case BPF_ALU|BPF_DIV|BPF_X:
			if (X == 0)
				return 0;
			A /= X;
			continue;
			
		case BPF_ALU|BPF_AND|BPF_X:
			A &= X;
			continue;
			
		case BPF_ALU|BPF_OR|BPF_X:
			A |= X;
			continue;

		case BPF_ALU|BPF_LSH|BPF_X:
			A <<= X;
			continue;

		case BPF_ALU|BPF_RSH|BPF_X:
			A >>= X;
			continue;

		case BPF_ALU|BPF_ADD|BPF_K:
			A += pc->k;
			continue;
			
		case BPF_ALU|BPF_SUB|BPF_K:
			A -= pc->k;
			continue;
			
		case BPF_ALU|BPF_MUL|BPF_K:
			A *= pc->k;
			continue;
			
		case BPF_ALU|BPF_DIV|BPF_K:
			A /= pc->k;
			continue;
			
		case BPF_ALU|BPF_AND|BPF_K:
			A &= pc->k;
			continue;
			
		case BPF_ALU|BPF_OR|BPF_K:
			A |= pc->k;
			continue;

		case BPF_ALU|BPF_LSH|BPF_K:
			A <<= pc->k;
			continue;

		case BPF_ALU|BPF_RSH|BPF_K:
			A >>= pc->k;
			continue;

		case BPF_ALU|BPF_NEG:
			A = -A;
			continue;

		case BPF_MISC|BPF_TAX:
			X = A;
			continue;

		case BPF_MISC|BPF_TXA:
			A = X;
			continue;
		}
	}
}

#ifdef _KERNEL
/*
 * Return true if the 'fcode' is a valid filter program.
 * The constraints are that each jump be forward and to a valid
 * code.  The code must terminate with either an accept or reject. 
 * 'valid' is an array for use by the routine (it must be at least
 * 'len' bytes long).  
 *
 * The kernel needs to be able to verify an application's filter code.
 * Otherwise, a bogus program could easily crash the system.
 */
int
bpf_validate(f, len)
	struct bpf_insn *f;
	int len;
{
	int i;
	struct bpf_insn *p;

	for (i = 0; i < len; ++i) {
		/*
		 * Check that that jumps are forward, and within 
		 * the code block.
		 */
		p = &f[i];
		if (BPF_CLASS(p->code) == BPF_JMP) {
			int from = i + 1;

			if (BPF_OP(p->code) == BPF_JA) {
				if (from + p->k >= (unsigned)len)
					return 0;
			}
			else if (from + p->jt >= len || from + p->jf >= len)
				return 0;
		}
		/*
		 * Check that memory operations use valid addresses.
		 */
		if ((BPF_CLASS(p->code) == BPF_ST ||
		     (BPF_CLASS(p->code) == BPF_LD && 
		      (p->code & 0xe0) == BPF_MEM)) &&
		    (p->k >= BPF_MEMWORDS || p->k < 0))
			return 0;
		/*
		 * Check for constant division by 0.
		 */
		if (p->code == (BPF_ALU|BPF_DIV|BPF_K) && p->k == 0)
			return 0;
	}
	return BPF_CLASS(f[len - 1].code) == BPF_RET;
}
#endif

#ifdef CONVERGED_IO
/*
 * This routine is invoked by the io subsystem when the echo
 * driver is configured as a converged io driver (i.e. has an
 * entry in the $DRIVER_INSTALL table in the master file).
 */
int 
bpfmod_install()
{
  int retval;

  if ((retval = str_install (&bpfmod_str_info)) != 0)
    {
      return (retval);
    }
  return 0;  /* return success */
}
#endif /* CONVERGED_IO */

/* This is called whenever we are pushed onto the stream. we need to
   allocate those  things generally found in the descriptor for
   BSD-BPF and we will probably use the module-private data area to
   store them. raj 10/98 */

bpfmodopen(q, dev, flag, sflag)
queue_t *q;
{
  struct bpf_d *d;

#ifdef BPFMOD_DEBUG
  printf("bpfmodopen: enter\n");
#endif /* BPFMOD_DEBUG */

  /* do we have enough memory in the system to allocate the
     descriptor? It would seem that the q ptr passed into the module 
     open routine is the read-side queue? raj 10/98 */

  MALLOC(d, struct bpf_d *, sizeof(struct bpf_d), M_STREAMS, M_WAITOK);

  if ((q->q_ptr = d) == NULL) {
    return(ENOMEM);
  }

  d->bd_bufsize = bpf_bufsize;
  d->bd_filter = NULL;
  d->bd_sbuf_mp = NULL;
  d->bd_immediate = 1;
  d->bd_rcount = 0;
  d->bd_dcount = 0;

  d->bd_timer_mp = NULL;
  d->bd_rtout = 0;
  d->bd_timer_id = 0;

  /* we will assume by default that we are operating over a DL_ETHER
     link, though for robustness, I suppose that pcap-bpfmod.c will
     need to worry about alignment */
  d->bd_hdrlen = BPF_WORDALIGN(14 + SIZEOF_BPF_HDR) - 14;

  /* we are using queue pair sync, and a single put routine, so we
     will make the write side pointer equal to the read-side pointer
     raj 10/98 */ 
  WR(q)->q_ptr = q->q_ptr;

  /* do we need to allocate some initial packet buffers here? raj 10/98 */
  return(0);
}

bpfmodclose(q)
queue_t *q;
{
  struct bpf_d *d = q->q_ptr;

#ifdef BPFMOD_DEBUG
  printf("bpfmodclose: enter\n");
#endif /* BPFMOD_DEBUG */

  /* if I have no service routine, I do not need to flush any queues
   raj 10/98 */ 

  /* we need to first see if there are any buffers to free, then we
     need to free the "descriptor" raj 10/98 */

  /* while we know that the bd_sbuf_mp is always a single mblk, it was
     suggested that freemsg is a good idea in case that changes in the
     future raj 10/98 */
  if (d->bd_sbuf_mp) freemsg(d->bd_sbuf_mp);

  /* if we were timing anything, clear that up too, first the timer,
     then the structures it might want to use. i start to wonder about
     race conditions...but decide to ignore them for the moment :) raj
     10/98 */

  if (d->bd_timer_id) untimeout(d->bd_timer_id);
  d->bd_timer_id = 0;
  if (d->bd_timer_mp) freemsg(d->bd_timer_mp);

  /* don't forget that we need to free the descriptor raj 10/98 */
  FREE(d,M_STREAMS);


  return(0);
}

/* Reset a descriptor by flushing its packet buffer and clearing the
   receive and drop counts.  raj 10/98 */

static void
reset_d(d)
	struct bpf_d *d;
{
	if (d->bd_sbuf_mp) {
	  freeb(d->bd_sbuf_mp);
	  d->bd_sbuf_mp = NULL;
	}

	d->bd_rcount = 0;
	d->bd_dcount = 0;
}

/* Set d's packet filter program to fp from the mblk bp.  If this file
   already has a filter,  free it and replace it.  Returns EINVAL for
   bogus requests.  
 */
int
bpf_setf(d, bp)
     struct bpf_d *d;
     mblk_t *bp;
{

  struct bpf_program *fp;
  struct bpf_insn *fcode, *old;
  u_int flen, size;
  int s;
  
#ifdef BPFMOD_DEBUG
  printf("bpf_setf: enter with d %p and bp %p\n",d,bp);
#endif /* BPFMOD_DEBUG */

  /* we currently do not handle programs in more than one mblk */
  if (bp->b_cont) {

#ifdef BPFMOD_DEBUG
    printf("bpf_setf: EINVAL too many mblks\n");
#endif /* BPFMOD_DEBUG */

    return(EINVAL);
  }

  old = d->bd_filter;

  fp = (struct bpf_program *)bp->b_rptr;

  /* if there are only four bytes in the mblk, that means that there
     were no instructions - what the BSD code checked by checking the
     fp->bf_insns pointer against 0. here in bpfmod however, the whole
     thing is contiguous. this means that trying to dereference
     fp->bf_insns could have unpredictable results since it really is
     not a pointer... raj 10/98 */
  if ((bp->b_wptr - bp->b_rptr) == 4) {
    if (fp->bf_len != 0) {

#ifdef BPFMOD_DEBUG
      printf("bpf_setf: EINVAL bf_len %d mblk len %d\n",
	     fp->bf_len,
	     bp->b_wptr - bp->b_rptr);
#endif /* BPFMOD_DEBUG */

      return (EINVAL);
    }

    d->bd_filter = NULL;
    reset_d(d);

    if (old != NULL)
      FREE((caddr_t)old, M_STREAMS);
    return (0);
  }
  flen = fp->bf_len;
  if (flen > BPF_MAXINSNS) {

#ifdef BPFMOD_DEBUG
    printf("bpf_setf: EINVAL flen %d > BPF_MAXINSNS\n",
	   flen);
#endif /* BPFMOD_DEBUG */

    return (EINVAL);
  }
  
  size = flen * sizeof(*fp->bf_insns);

  /* this may be a duplicate check since we've already looked for a
     b_cont, but I'm not terribly confident of what I'm doing here
     just now, so we'll add the check. i really don't like
     ioctls... raj 10/98 */ 

  if (size > ((bp->b_wptr - bp->b_rptr) - 4)) {

#ifdef BPFMOD_DEBUG
    printf("bpf_setf: EINVAL size %d > mblk len - 4 %d\n",
	   size,
	   bp->b_wptr - bp->b_rptr);
#endif /* BPFMOD_DEBUG */

    return(EINVAL);
  }

  MALLOC(fcode, struct bpf_insn *, size, M_STREAMS, M_NOWAIT);

  if (fcode == NULL)
    return (ENOMEM);

  /* the copyin in the original is clearly bogus for the streams
     module version as the copyin will have happened at the stream
     head, however, since the whole thing is contiguous, we start
     sizeof(fp->bf_len) past the beginning, and pray :) raj 10/98 */

#ifdef BPFMOD_DEBUG
  printf("bpf_setf: copy %d program bytes from %p to %p orig %p\n",
	 size,
	 (caddr_t)(fp)+sizeof(fp->bf_len),
	 fcode,
	 (caddr_t)(fp+sizeof(fp->bf_len)));
  printf("bpf_setf: fp %p\n",fp);
#endif /* BPFMOD_DEBUG */

  bcopy((caddr_t)(fp)+sizeof(fp->bf_len),(caddr_t)fcode,size);

#ifdef BPFMOD_DEBUG
  {
    int i;
    u_char *bc = (u_char *)fcode;

    printf("fcode from %p",bc);
    for(i=0;i<size;i++) {
      if ((i % 8) == 0) printf("\n\t");
      printf("%2.2x ",*bc++);
    }
    printf("\n");
  }
#endif /* BPFMOD_DEBUG */

  if (bpf_validate(fcode, (int)flen)) {

    d->bd_filter = fcode;
    reset_d(d);

    if (old != NULL)
      FREE((caddr_t)old, M_STREAMS);
    
    return (0);
  }
  FREE((caddr_t)fcode, M_STREAMS);

#ifdef BPFMOD_DEBUG
  printf("bpf_setf: EINVAL bpf_validate failure\n");
#endif /* BPFMOD_DEBUG */

  return (EINVAL);
}

/*
  Name: bpfmod_timeout_int(q, mp)
  Description: If there is a store buffer with bytes in it, send it up
  the stream. Otherwise, return having done nothing. This code is
  intended to "emulate" the read timeout feature of the original BSD
  BPF so when the store buffer is large, one does not wait forever to
  see the packets. we assume that the queue we are called with is the
  "read side" queue so a putnext goes the proper direction lest
  someone get confused raj 10/98 */

bpfmod_timeout_int(queue_t *q, mblk_t *mp)
{
  struct bpf_d *d = q->q_ptr;
  mblk_t *rp;

  if (d->bd_sbuf_mp && 
      (d->bd_sbuf_mp->b_rptr != d->bd_sbuf_mp->b_wptr)) {

    /* we have a store buffer with data, now can we get the header
       mblk? i'd love to determine once and for all if I really need
       to sent M_PROTO's up, or if I can just send M_DATAs and avoice
       this extra allocb. raj 10/98 */

    if ((rp = allocb(sizeof(dl_hp_rawdata_ind_t),BPRI_LO)) != NULL) {
      rp->b_wptr = rp->b_rptr + sizeof(dl_hp_rawdata_ind_t);
      ((dl_hp_rawdata_ind_t *)rp)->dl_primitive = DL_HP_RAWDATA_IND;
      rp->b_cont = d->bd_sbuf_mp;

      putnext(q,d->bd_sbuf_mp);
      /* we no longer have a buffer, set things up so we try to get one
	 next time a packet arrives raj 10/98 */
      d->bd_sbuf_mp = NULL;
      d->bd_slen = 0;
    }
  }      
}

/*
  Name bpfmod_timeout(caddr_t arg)
  Description: This is the routine called by the systems timer
  facility. The value of arg is expected to point at a queue_t. The
  queue_t->q_ptr will point at a bpf_d, which will in turn point to a
  bd_timer_mp which will be used by the streams_put() call to get us
  "into" the streams framework. raj 10/98 */

bpfmod_timeout(caddr_t arg)
{
  queue_t *q = (queue_t *)arg;
  struct bpf_d *d = q->q_ptr;

  streams_put((streams_put_t)bpfmod_timeout_int,
	      q,
	      d->bd_timer_mp,
	      q); /* passing a NULL private pointer to
		     bpfmod_timeout_int is a nono, and the "q" pointer
		     used to find the queue is *not* passed, the
		     private pointer is... the streams manual needs to
		     be much more explicit about that one... raj 10/98
		     */

  /* don't forget to flag that our timer is not running any longer, so
     catcher will know to start one again once another packet comes
     in. raj 10/98 */
  d->bd_timer_id = 0;

}

/*
  Name: copybn(mp, dest, len)
  Description: Copy len bytes from the message mp to the destination
  dest. It is assumed that mp and dest have at least len bytes.
  raj 10/98 */

copybn(mblk_t *mp, caddr_t dest, int len)
{
  int bytes_remaining = len;
  int bytes_to_copy;
  mblk_t *mp1 = mp;
  caddr_t tmp=dest;

#ifdef BPFMOD_DEBUG
  printf("copybn: request copy %d bytes from mblk %p to buffer at %p\n",
	 len, mp, dest);
#endif /* BPFMOD_DEBUG */


  while (bytes_remaining) {
    bytes_to_copy = min(bytes_remaining,(mp1->b_wptr - mp1->b_rptr));
    bcopy((caddr_t)mp1->b_rptr,(caddr_t)tmp,bytes_to_copy);

    /* silly programmer, don't you realize you have to update the
       destination pointer too!?! raj 10/98 */
    tmp += bytes_to_copy;

    bytes_remaining -= bytes_to_copy;
    
    mp1 = mp1->b_cont;
  }
}

/* Name: catchpacket(q, bp, pktlen, snaplen)
   Description: Copy min(pktlen,snaplen) bytes from the packet pointed
   at by bp into the store buffer associated with our queue q. If we
   are in immediate mode, or are about to overflow the store buffer,
   send the current store buffer up the stream. if we cannot allocated
   space to store the desirec bytes, increment the drop counts. this
   isn't exactly written in the best way for PA in terms of
   performance, so at some point the flow should really be
   cleaned-up. 
 raj 10/98 */

catchpacket(q, bp, pktlen, snaplen)
     queue_t *q;
     mblk_t *bp;
     u_int pktlen, snaplen;
{
  struct bpf_d *d = q->q_ptr;
  struct bpf_hdr *hp;
  int totlen, curlen;
  int hdrlen = d->bd_hdrlen;
  mblk_t *rp; /* this will hold the "DL_HP_RAWDATA_IND" */

#ifdef BPFMOD_DEBUG
  printf("catchpacket: q %p, bp %p, pktlen %d, snaplen %d\n",
	 q,bp,pktlen,snaplen);
#endif /* BPFMOD_DEBUG */

  /* we need to check that we indeed have a store buffer, and if we do
   not, to get one raj 10/98 */
  if (d->bd_sbuf_mp == NULL) {
    if ((d->bd_sbuf_mp = allocb(d->bd_bufsize,BPRI_LO)) == NULL) {
      /* we cannot get any place to put the packet, so we will just
	 drop the thing upon return to the caller raj 10/98 */
      ++d->bd_dcount;
      return;
    }

    d->bd_slen = 0;
#ifdef BPFMOD_DEBUG
  printf("catchpacket: allocated store mblk %p buffer %p from point 1\n",
	 d->bd_sbuf_mp,d->bd_sbuf_mp->b_wptr);
#endif /* BPFMOD_DEBUG */

  
  }

  /* Figure out how many bytes to move.  If the packet is greater or
    equal to the snapshot length, transfer that much.  Otherwise,
    transfer the whole packet (unless we hit the buffer size limit).  raj 10/98 */

  totlen = hdrlen + MIN(snaplen, pktlen);
  if (totlen > d->bd_bufsize)
    totlen = d->bd_bufsize;

  /* Round up the end of the previous packet to the next longword. if
   we do not, the system might get cranky when we try to overlay the
   bpf encapsulation header. I am pretty sure that the lentgh of the
   data at this point is always word aligned, but I guess it is OK to
   be paranoid. raj 10/98 */

  curlen = BPF_WORDALIGN(d->bd_slen);
  if (curlen + totlen > d->bd_bufsize) {

    /* This packet will overflow the storage buffer. Send what we have
       up the stream (our analogue to the wakeup), get a new buffer if
       we can (a la rotate buffers), otherwise, drop the packet and
       increment the appropriate stat. question - should I be doing a
       canput here? i don't know. another good question is whether or
       not I should be pre-pending an M_PROTO here with something like
       DL_HP_BPFDATA_IND... raj 10/98 */

#ifdef BPFMOD_DEBUG
    printf("catchpacket: pushing mblk %p to stream head from point 1\n",
	   d->bd_sbuf_mp);
    printf("             rptr %p wptr %p\n",
	   d->bd_sbuf_mp->b_rptr,d->bd_sbuf_mp->b_wptr);
#endif /* BPFMOD_DEBUG */

    /* this does look a little squirrely, but I'm still figuring-out
       just what the right way to do the messages up to the stream
       head so the user does not get a pair of M_DATA's merged - given
       the way the bpf header works, that could lead to some strange
       results. raj 10/98 */
  
    if ((rp = allocb(sizeof(dl_hp_rawdata_ind_t),BPRI_LO)) == NULL) {
      /* we cannot allocate the buffer for the DL_HP_RAWDATA_IND, so
	 we will just drop the current packet perhaps the next time a
	 packet comes-in, we will be able to get a haeder for our
	 store buffer */
      ++d->bd_dcount;
      return;
    }
    else {
      /* we could get the mblk for the data indicator */

      rp->b_wptr = rp->b_rptr + sizeof(dl_hp_rawdata_ind_t);
      ((dl_hp_rawdata_ind_t *)rp)->dl_primitive = DL_HP_RAWDATA_IND;
      rp->b_cont = d->bd_sbuf_mp;
      
      putnext(q,d->bd_sbuf_mp);

      if ((d->bd_sbuf_mp = allocb(d->bd_bufsize,BPRI_LO)) == NULL) {
	/* we cannot get any place to put the packet, so we will just
	   drop the thing upon return to the caller raj 10/98 */
	++d->bd_dcount;
	return;
	
      }
    }
    
    /* at this point we have a fresh buffer for the packet. there are
       probably other fields of the mblk that need to be set
       correctly, but I'll worry about that later :) Later I also
       should probably worry about whether or not the rptr and wptr's
       of a newly allocated mblk arrive "aligned" to a four-byte
       boundary... raj 10/98 */

    d->bd_slen = 0;
    curlen = 0;

#ifdef BPFMOD_DEBUG
  printf("catchpacket: allocated store mblk %p buffer %p from point 2\n",
	 d->bd_sbuf_mp,d->bd_sbuf_mp->b_wptr);
#endif /* BPFMOD_DEBUG */


  }

  /* Append the bpf header. We better have made sure we keep the
     b_wptr updated and properly aligned, or the machine might get
     rather cranky. at this point, curlen has been rounded-up to the
     next word boundary, so everything should be fine raj 10/98 */

  d->bd_sbuf_mp->b_wptr = d->bd_sbuf_mp->b_rptr + curlen;
  hp = (struct bpf_hdr *)(d->bd_sbuf_mp->b_wptr);

  hp->bh_tstamp = ms_gettimeofday();
  hp->bh_datalen = pktlen;
  hp->bh_hdrlen = hdrlen;
  hp->bh_caplen = totlen - hdrlen;

  /* since we've stuck a bpf header here, update wptr so we don't copy
     over the header. */

  d->bd_sbuf_mp->b_wptr += hdrlen;

  /* Copy the packet data into the store buffer */

  copybn(bp, 
	 (caddr_t)d->bd_sbuf_mp->b_wptr, 
	 hp->bh_caplen);

  /* we will worry about alignment of the b_wptr before we go to write
     to it again. if we fix it here, we end-up making our buffer
     arbitrarily large, which is something the BPF-sin app above us is
     not expecting. raj 10/98 */

  d->bd_sbuf_mp->b_wptr += hp->bh_caplen;
  d->bd_slen = curlen + totlen;

  /* by now, we have either copied the data, or bolted, so if we are
     in immediate mode, send the buffer up the stream raj 10/98 */ 

  if (d->bd_immediate) {

#ifdef BPFMOD_DEBUG
    printf("catchpacket: pushing mblk %p to stream head from point 2\n",
	   d->bd_sbuf_mp);
    printf("             rptr %p wptr %p\n",
	   d->bd_sbuf_mp->b_rptr,d->bd_sbuf_mp->b_wptr);
#endif /* BPFMOD_DEBUG */

    /* if we could not get the mblk for the data ind, then we'll just
       let the current buffer sit here, even though we are in
       immediate mode. */
    if((rp = allocb(sizeof(dl_hp_rawdata_ind_t),BPRI_LO)) != NULL) {
      rp->b_wptr = rp->b_rptr + sizeof(dl_hp_rawdata_ind_t);
      ((dl_hp_rawdata_ind_t *)rp)->dl_primitive = DL_HP_RAWDATA_IND;
      rp->b_cont = d->bd_sbuf_mp;

      putnext(q,d->bd_sbuf_mp);
      /* we no longer have a buffer, set things up so we try to get one
	 next time a packet arrives raj 10/98 */
      d->bd_sbuf_mp = NULL;
      d->bd_slen = 0;
    }
  }
  else {
    /* we are not in immediate mode, so see if we should enable a
       timer to push packets up the stream. just be sure to not enable
       a timer if we have one already running, that could be bad raj 10/98 */

#ifdef BPFMOD_DEBUG
      printf("catchpacket: pre timeout bd_timer_id %d bd_rtout %d\n", 
	     d->bd_timer_id,d->bd_rtout);
#endif /* BPFMOD_DEBUG */

    if ((d->bd_rtout != 0) &&
	(d->bd_timer_id ==0 )) {
      d->bd_timer_id = timeout(bpfmod_timeout,
			       (caddr_t) q,
			       d->bd_rtout);

#ifdef BPFMOD_DEBUG
      printf("catchpacket: post timeout d->bd_timer_id %d\n",d->bd_timer_id);
#endif /* BPFMOD_DEBUG */

    }
  }

}

/*
  Name: bpfmodrput(q,bp)
  Description:
  Depending on the message type, enact processing or pass through without
  any processing.
  raj 10/98 */

bpfmodrput(q, bp)
queue_t *q;
mblk_t *bp;
{

  struct bpf_d *d = q->q_ptr;
  u_int pktlen,slen;

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: enter q = %p d = %d bp = %p\n",q,d,bp);
#endif /* BPFMOD_DEBUG */

  switch(bp->b_datap->db_type) {

  case M_PROTO: {
    dl_hp_rawdata_ind_t *rd_ind = (dl_hp_rawdata_ind_t *)bp->b_rptr;

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: M_PROTO\n");
#endif /* BPFMOD_DEBUG */


    /* we probably should be making sure there is enough data in the
       mblk for a DL_HP_RAWDATA_IND. i'm sure that one of these days,
       not checking will result in a data segmentation violation or
       something, however, until that time, lets not put any more
       kruft into the data path than we have to. raj 10/98 */
    
    if (rd_ind->dl_primitive == DL_HP_RAWDATA_IND) {

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: DL_HP_RAWDATA_IN\n");
#endif /* BPFMOD_DEBUG */

      /* we have a raw mode packet, hit it with the filter. yes,
	 msgdsize is ugly, but until folks start stashing the packet
	 size in a control field of the mblk someplace... I am
	 assuming that the M_PROTO will not be counted. I guess I
	 could use bp->b_cont but it should not be necessary. raj 10/98 */

      pktlen = msgdsize(bp->b_cont);

      /* In the original bpf_tap/bpf_mtap code, this is increamented
	 on each call to the routine - of course, there, every call to
	 bpf_tap is for a packet. bpfmodrput will sometimes be called
	 for things that are not packets for filtering, so we will
	 only increment the stat when we are going to call the filter
	 routine raj 10/98 */

      ++d->bd_rcount;

      /* I cannot fathom how one could ever have an
	 M_PROTO/DL_HP_RAWDATA_IND with a null b_cont so I am not
	 going to check for it. famous last words raj 10/98 */

      if ((slen = bpf_filter(d->bd_filter, bp->b_cont, pktlen, 0)) != 0) {
	/* the filter would like us to snarf this packet */

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: snarfing %d bytes of %d byte packet at %p\n",
	   slen,pktlen, bp->b_cont);
#endif /* BPFMOD_DEBUG */

	catchpacket(q, bp->b_cont, pktlen, slen);
      }

      /* at this point, either we wanted the packet and had space for
	 it, or wanted the packet but did not have the space, or did
	 not want the packet. In each case we can just free the
	 message. I'm pretty sure the implicit return is sufficient
	 here raj 10/98 */

      freemsg(bp);

    }
    else {

      /* we do not know exactly what is being passed-up, which means
	 we should just pass it along and trust that someone else
	 knows how to deal with it. this may confuse the poor
	 bpf-expecting app on the other end, but then, this situation
	 is probably his fault in the first place. the only other
	 option would be to pass-up an M_ERROR message, which in the
	 end may be the right thing to do after all, but I have my
	 doubts. raj 10/98 */

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: unknown M_PROTO\n");
#endif /* BPFMOD_DEBUG */


      putnext(q, bp);

    }
    break;
  }
  case M_PCPROTO:

    /* this might be one of the DLPI primitives we wish to wiretap as
       part of our simulating BPF. at some point, there will be state
       information we need to consult to see if the M_PCPROTO is one
       generated by us so we know if it should be passed-up. for now,
       we are completely passive and so always pass the beast back up
       raj 10/98 */ 

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: M_PCPROTO\n");
#endif /* BPFMOD_DEBUG */


    {
      dl_info_ack_t *dl_info_ack = (dl_info_ack_t *)bp->b_rptr;
      
      /* at some point, I should probably be checking the
	 dl_current_state because the dl_mac_type is only valid when
	 we are attached raj 10/98 */

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: DL_INFO_ACK\n");
#endif /* BPFMOD_DEBUG */


      if (dl_info_ack->dl_primitive = DL_INFO_ACK) {
	switch (dl_info_ack->dl_mac_type) {
	case DL_ETHER:
	  d->bd_dlt = DLT_EN10MB;
	  d->bd_hdrlen = BPF_WORDALIGN(14 + SIZEOF_BPF_HDR) - 14;

#ifdef BPFMOD_DEBUG
	  printf("bpfmodrput: DL_ETHER mac, hdrlen now %d\n",
		 d->bd_hdrlen);
#endif /* BPFMOD_DEBUF */

	  break;
	case DL_FDDI:
	  d->bd_dlt = DLT_FDDI;
	  d->bd_hdrlen = BPF_WORDALIGN(21 + SIZEOF_BPF_HDR) - 21;

#ifdef BPFMOD_DEBUG
	  printf("bpfmodrput: DL_FDDI mac, hdrlen now %d\n",
		 d->bd_hdrlen);
#endif /* BPFMOD_DEBUF */

	  break;
	case DL_CSMACD:
	  /* for some reason, the data link type flips from DL_ETHER
	     to DL_CSMACD after one of pcap-bpfmod's requests,
	     however, the chances that we are really just receiveing 
	  d->bd_dlt = DLT_ETHER;
	  d->bd_hdrlen = BPF_WORDALIGN(14 + SIZEOF_BPF_HDR) - 14;

#ifdef BPFMOD_DEBUG
	  printf("bpfmodrput: DL_CSMACD mac, assume DL_ETHER hdrlen now %d\n",
		 d->bd_hdrlen);
#endif /* BPFMOD_DEBUF */

	  break;
	default:
	  d->bd_dlt = DLT_NULL;
	  d->bd_hdrlen = BPF_WORDALIGN(SIZEOF_BPF_HDR);

#ifdef BPFMOD_DEBUG
	  printf("bpfmodrput: unfamiliar mac, hdrlen now %d\n",
		 d->bd_hdrlen);
#endif /* BPFMOD_DEBUF */

	  break;
	}
      }
      putnext(q,bp);
      break;
    }
  default:

    /* we are just a module without a service routine, so we can just
       putnext anything we do not understand. that way, if someone for
       some inexplicable reason pushes bpfmod into a place it is not
       intended, data will still flow. for something like libpcap, the
       proper order is to open /dev/dlpi, push bpfmod, set the PPA,
       set raw mode, program the filter and *then* and only then
       enable promiscuous mode, otherwise, the app may not get bpf
       messages at first  - at least I think that is the proper order
       :) raj 10/98 */

#ifdef BPFMOD_DEBUG
    printf("bpfmodrput: db_type %d not handled\n",bp->b_datap->db_type);
#endif /* BPFMOD_DEBUG */


    putnext(q, bp);
    break;
  }
}

/*
  Name: bpfmodioctl(q,bp)
  Description:
  Depending on the ioctl type, either process it, or pass it along
  without any processing. raj 10/98
  */

bpfmodioctl(q, bp)
     queue_t *q;
     mblk_t *bp;

{
  struct iocblk *iocp;
  struct bpf_d *d = q->q_ptr;

#ifdef BPFMOD_DEBUG
    printf("bpfmodioctl: enter q = %p bp = %p\n",q,bp);
#endif /* BPFMOD_DEBUG */

    iocp = (struct iocblk *) bp->b_rptr;

    switch (iocp->ioc_cmd) {

    case FIONREAD:

      /* Check for read packet available, I'm not sure we will ever
	 see this ioctl below the stream head however, so we may need
	 to make a BFIONREAD or some such, once I figure-out how to
	 pick ioctl numbers...  raj 10/98 */  

      {

	/* we may not actually have a buffer at the moment, so best
	   check, or we will chase a NULL pointer off into the great
	   beyond. raj 10/98 */
	if (d->bd_sbuf_mp) {
	  iocp->ioc_rval = 
	    d->bd_sbuf_mp->b_wptr -  d->bd_sbuf_mp->b_rptr;
	}
	else {
	  iocp->ioc_rval = 0;
	}

	iocp->ioc_error = 0;
	iocp->ioc_count = 0;
	bp->b_datap->db_type = M_IOCACK;
	qreply(q,bp);
	break;
      }

    case BIOCGBLEN:
    /* Get buffer len [for read()]. */
      if ((bp->b_cont == NULL) || 
	  ((bp->b_cont->b_wptr - bp->b_cont->b_rptr) != sizeof(int)) ||
	  (iocp->ioc_count != sizeof(int))) {
	/* there is supposed to be a value present in another mblk
	   chained to the M_IOCTL mblk and that mblk should be the
	   size of an int raj 10/98 */
	bp->b_datap->db_type = M_IOCNAK;
	iocp->ioc_error = EINVAL;
#ifdef BPFMOD_DEBUG
	printf("bpfmodioctl: BIOCGBLEN EINVAL\n");
#endif /* BPFMOD_DEBUG */
      }
      else {

	*(u_int *)(bp->b_cont->b_rptr) = d->bd_bufsize;
      
	bp->b_datap->db_type = M_IOCACK;

#ifdef BPFMOD_DEBUG
	printf("bpfmodioctl: BIOCGBLEN to %d\n",d->bd_bufsize);
#endif /* BPFMOD_DEBUG */

      }

      qreply(q,bp);
      break;
    
    case BIOCSBLEN:

    /* Set buffer length. */

      if ((bp->b_cont == NULL) || 
	  ((bp->b_cont->b_wptr - bp->b_cont->b_rptr) != sizeof(int)) ||
	  (iocp->ioc_count != sizeof(int))) {
	/* there is supposed to be a value present in another mblk
	   chained to the M_IOCTL mblk and that mblk should be the
	   size of an int raj 10/98 */
	bp->b_datap->db_type = M_IOCNAK;
	iocp->ioc_error = EINVAL;
#ifdef BPFMOD_DEBUG
	printf("bpfmodioctl: BIOCSBLEN EINVAL\n");
#endif /* BPFMOD_DEBUG */
      }
      else {

	u_int size = *(u_int *)(bp->b_cont->b_rptr);
      
	bp->b_datap->db_type = M_IOCACK;

	if (size > BPF_MAXBUFSIZE)
	  size = BPF_MAXBUFSIZE;
	else if (size < BPF_MINBUFSIZE)
	  size = BPF_MINBUFSIZE;
	d->bd_bufsize = size;

#ifdef BPFMOD_DEBUG
	printf("bpfmodioctl: BIOCSBLEN to %d\n",size);
#endif /* BPFMOD_DEBUG */

      }

      qreply(q,bp);
      break;

    case BIOCSETF: 

#ifdef BPFMOD_DEBUG
      printf("bpfioctl: BIOCSETF\n");
#endif /* BPFMOD_DEBUG */

      /* Set link layer read filter. I am reasonably confident that a
	 BPF program will fit within a single mblk - more of those
	 famous last words :) raj 10/98 */

      if (bp->b_cont == NULL) {
	/* is this correct behaviour? I'm wondering if instead the
	   filter pointer in the descriptor should just be set to
	   NULL? raj 10/98 */
	iocp->ioc_error = EINVAL;
	bp->b_datap->db_type = M_IOCNAK;
      }
      else {
	iocp->ioc_error = bpf_setf(d, 
				   (struct bpf_program *)bp->b_cont);

	if (iocp->ioc_error != 0) bp->b_datap->db_type = M_IOCNAK;
	else bp->b_datap->db_type = M_IOCACK;
      }

      iocp->ioc_count = 0;
      qreply(q,bp);
      break;
    

    case BIOCFLUSH:

#ifdef BPFMOD_DEBUG
      printf("bpfmodioctl: BIOCFLUSH\n");
#endif /* BPFMOD_DEBUG */

      reset_d(d);
      bp->b_datap->db_type = M_IOCACK;
      qreply(q,bp);

      break;
    
    case BIOCGDLT:

#ifdef BPFMOD_DEBUG
      printf("bpfmodioctl: BIOCGDLT\n");
#endif /* BPFMOD_DEBUG */

      /* return the current setting of the data link type in BPF
	 terms. presently, we rely on having seen a DL_INFO_ACK go by
	 in bpfmodwput before this. at some point we may change this
	 so we send a DL_INFO_REQ down upon receipt of this request,
	 and then reply once we see a DL_INFO_ACK go by. however, that
	 will require a bit more of a state machine than I wish to
	 provide just at the moment raj 10/98 */

      iocp->ioc_rval = d->bd_dlt;
      iocp->ioc_count = 0;
      bp->b_datap->db_type = M_IOCACK;
      qreply(q,bp);
      break;

    case BIOCSRTIMEOUT:
      {
	struct timeval *tv;
	bpf_u_int32 msec;
	if ((bp->b_cont) && 
	    (iocp->ioc_count >= sizeof(struct timeval)) &&
	    ((d->bd_timer_mp = allocb(1,BPRI_LO)) != NULL)) {
	  tv = (struct timeval *)bp->b_cont->b_rptr;

	  /* figure-out the number of ticks in the timeval */
	  msec = (tv->tv_sec * 1000) + (tv->tv_usec / 1000);
	  d->bd_rtout = msec / (CLOCKS_PER_SEC / 1000);

	  if (d->bd_rtout == 0 && tv->tv_usec != 0)
				d->bd_rtout = 1;

#ifdef BPFMOD_DEBUG
	  printf("bpfmodioctl: BIOCSRTIMEOUT set to %d ticks\n",
		 d->bd_rtout);
#endif /* BPFMOD_TIMEOUT */

	  iocp->ioc_error = 0;
	  iocp->ioc_count = 0;
	  bp->b_datap->db_type = M_IOCACK;
	}
	else {
	  /* something was bogus */

#ifdef BPFMOD_DEBUG
	  printf("bpfmodioctl: BIOCSRTIMEOUT EINVAL bcont %p count %d mp %p\n",
		 bp->b_cont,
		 iocp->ioc_count,
		 d->bd_timer_mp);
#endif /* BPFMOD_TIMEOUT */

	  iocp->ioc_error = EINVAL;
	  bp->b_datap->db_type = M_IOCNAK;
	}
	qreply(q,bp);
	break;
      }

    case BIOCPROMISC:
    case BIOCGETIF:
    case BIOCSETIF:
    case BIOCGRTIMEOUT:
      iocp->ioc_error = ENOTSUP;
      iocp->ioc_count = 0;
      bp->b_datap->db_type = M_IOCNAK;
      qreply(q,bp);
      break;
  
    case BIOCGSTATS:
	  {
	    struct bpf_stat *bs;

	    if (bp->b_cont == NULL) {
	      iocp->ioc_error = EINVAL;
	      bp->b_datap->db_type = M_IOCNAK;

#ifdef BPFMOD_DEBUG
    printf("bpfmodioctl: BIOCGSTATS EINVAL\n");
#endif /* BPFMOD_DEBUG */

	    }
	    else {
	      bs = (struct bpf_stat *)(bp->b_cont->b_rptr);
	      bs->bs_recv = d->bd_rcount;
	      bs->bs_drop = d->bd_dcount;
	      bp->b_datap->db_type = M_IOCACK;

#ifdef BPFMOD_DEBUG
    printf("bpfmodioctl: BIOCGSTATS recv %d drop %d\n",
	   d->bd_rcount,
	   d->bd_dcount);
#endif /* BPFMOD_DEBUG */

	    }
	    qreply(q,bp);
	    break;
	  }
  
    case BIOCIMMEDIATE:
      if (bp->b_cont == NULL) {
	iocp->ioc_error = EINVAL;
	bp->b_datap->db_type = M_IOCNAK;

#ifdef BPFMOD_DEBUG
    printf("bpfmodioctl: BIOCIMMEDIATE EINVAL\n");
#endif /* BPFMOD_DEBUG */

      }
      else {
	d->bd_immediate = *(u_int *)(bp->b_cont->b_rptr);
	bp->b_datap->db_type = M_IOCACK;

#ifdef BPFMOD_DEBUG
    printf("bpfmodioctl: BIOCIMMEDIATE to %d\n",d->bd_immediate);
#endif /* BPFMOD_DEBUG */

      }
      qreply(q,bp);
      break;
    
    case BIOCVERSION:
      {
	struct bpf_version *bv;

	if (bp->b_cont == NULL) {
	  iocp->ioc_error = EINVAL;
	  bp->b_datap->db_type = M_IOCNAK;

#ifdef BPFMOD_DEBUG
    printf("bpfmodioctl: BIOCVERSION EINVAL\n");
#endif /* BPFMOD_DEBUG */

	}
	else {
	  bv = (struct bpf_version *)(bp->b_cont->b_rptr);

	  bv->bv_major = BPF_MAJOR_VERSION;
	  bv->bv_minor = BPF_MINOR_VERSION;
	  bp->b_datap->db_type = M_IOCACK;

#ifdef BPFMOD_DEBUG
    printf("bpfmodioctl: BIOCVERSION check\n");
#endif /* BPFMOD_DEBUG */

	}
	qreply(q,bp);
	break;
      }

    default:
      /* if we do not know what it is, just pass it along */
      putnext(q,bp);
      break;
    }

}



/* 
   Name: bpfmodwput(q,bp)
   This routing implements the "write" side of the bpfmod - we only
   intercept and process M_IOCTLs to control the filters and
   such. M_DATAs and such will just be passed directly on to the next
   module (most likely the DLPI driver). It would seem that since we
   have no service routine, we do not need to process M_FLUSH
   messages. 

   it is the read side (bpfmodrput) which will filter the messages
   containing inbound packets.
   raj 10/98 */

bpfmodwput(q, bp)
queue_t *q;
mblk_t *bp;
{


#ifdef BPFMOD_DEBUG
    printf("bpfmodwput: enter q = %p bp = %p\n",q,bp);
#endif /* BPFMOD_DEBUG */

  switch(bp->b_datap->db_type) {

  case M_IOCTL: 
    bpfmodioctl(q,bp);
    break;

  default:
    /* it was not a message type we were expecting to process, so just
       send it along its way raj 10/98 */
    putnext(q, bp);
    break;
  }
}

/* do I even need or want a service routine? I think the answer is no,
 so lets not have one raj 10/98 */
